@@ -29,79 +29,33 @@ see https://www.gnu.org/licenses/. */
 #ifndef PAGMO_ALGORITHM_HPP
 #define PAGMO_ALGORITHM_HPP
 
+#include <cassert>
 #include <memory>
+#include <string>
 #include <type_traits>
 #include <typeinfo>
 #include <utility>
 
+#include <boost/type_traits/integral_constant.hpp>
+#include <boost/type_traits/is_virtual_base_of.hpp>
+
 #include <pagmo/detail/make_unique.hpp>
+#include <pagmo/detail/visibility.hpp>
 #include <pagmo/exceptions.hpp>
 #include <pagmo/population.hpp>
-#include <pagmo/serialization.hpp>
+#include <pagmo/s11n.hpp>
 #include <pagmo/threading.hpp>
 #include <pagmo/type_traits.hpp>
 
-/// Macro for the registration of the serialization functionality for user-defined algorithms.
-/**
- * This macro should always be invoked after the declaration of a user-defined algorithm: it will register
- * the algorithm with pagmo's serialization machinery. The macro should be called in the root namespace
- * and using the fully qualified name of the algorithm to be registered. For example:
- * @code{.unparsed}
- * namespace my_namespace
- * {
- *
- * class my_algorithm
- * {
- *    // ...
- * };
- *
- * }
- *
- * PAGMO_REGISTER_ALGORITHM(my_namespace::my_algorithm)
- * @endcode
- */
-#define PAGMO_REGISTER_ALGORITHM(algo) CEREAL_REGISTER_TYPE_WITH_NAME(pagmo::detail::algo_inner<algo>, "uda " #algo)
+#define PAGMO_S11N_ALGORITHM_EXPORT_KEY(algo)                                                                          \
+    BOOST_CLASS_EXPORT_KEY2(pagmo::detail::algo_inner<algo>, "uda " #algo)                                             \
+    BOOST_CLASS_TRACKING(pagmo::detail::algo_inner<algo>, boost::serialization::track_never)
 
-namespace pagmo
-{
+#define PAGMO_S11N_ALGORITHM_IMPLEMENT(algo) BOOST_CLASS_EXPORT_IMPLEMENT(pagmo::detail::algo_inner<algo>)
 
-/// Null algorithm
-/**
- * This algorithm is used to implement the default constructors of pagmo::algorithm and of the meta-algorithms.
- */
-struct null_algorithm {
-    /// Evolve method.
-    /**
-     * In the null algorithm, the evolve method just returns the input
-     * population.
-     *
-     * @param pop input population.
-     *
-     * @return a copy of the input population.
-     */
-    population evolve(const population &pop) const
-    {
-        return pop;
-    };
-    /// Algorithm name.
-    /**
-     * @return <tt>"Null algorithm"</tt>.
-     */
-    std::string get_name() const
-    {
-        return "Null algorithm";
-    }
-    /// Serialization support.
-    /**
-     * This class is stateless, no data will be loaded or saved during serialization.
-     */
-    template <typename Archive>
-    void serialize(Archive &)
-    {
-    }
-};
-
-} // namespace pagmo
+#define PAGMO_S11N_ALGORITHM_EXPORT(algo)                                                                              \
+    PAGMO_S11N_ALGORITHM_EXPORT_KEY(algo)                                                                              \
+    PAGMO_S11N_ALGORITHM_IMPLEMENT(algo)
 
 namespace pagmo
 {
@@ -204,10 +158,10 @@ template <typename T>
 class is_uda
 {
     static const bool implementation_defined
-        = (std::is_same<T, uncvref_t<T>>::value && std::is_default_constructible<T>::value
-           && std::is_copy_constructible<T>::value && std::is_move_constructible<T>::value
-           && std::is_destructible<T>::value && has_evolve<T>::value)
-          || detail::disable_uda_checks<T>::value;
+        = detail::disjunction<detail::conjunction<std::is_same<T, uncvref_t<T>>, std::is_default_constructible<T>,
+                                                  std::is_copy_constructible<T>, std::is_move_constructible<T>,
+                                                  std::is_destructible<T>, has_evolve<T>>,
+                              detail::disable_uda_checks<T>>::value;
 
 public:
     /// Value of the type trait.
@@ -220,7 +174,7 @@ const bool is_uda<T>::value;
 namespace detail
 {
 
-struct algo_inner_base {
+struct PAGMO_DLL_PUBLIC_INLINE_CLASS algo_inner_base {
     virtual ~algo_inner_base() {}
     virtual std::unique_ptr<algo_inner_base> clone() const = 0;
     virtual population evolve(const population &pop) const = 0;
@@ -232,13 +186,13 @@ struct algo_inner_base {
     virtual std::string get_extra_info() const = 0;
     virtual thread_safety get_thread_safety() const = 0;
     template <typename Archive>
-    void serialize(Archive &)
+    void serialize(Archive &, unsigned)
     {
     }
 };
 
 template <typename T>
-struct algo_inner final : algo_inner_base {
+struct PAGMO_DLL_PUBLIC_INLINE_CLASS algo_inner final : algo_inner_base {
     // We just need the def ctor, delete everything else.
     algo_inner() = default;
     algo_inner(const algo_inner &) = delete;
@@ -251,7 +205,7 @@ struct algo_inner final : algo_inner_base {
     // The clone method, used in the copy constructor of algorithm.
     virtual std::unique_ptr<algo_inner_base> clone() const override final
     {
-        return make_unique<algo_inner>(m_value);
+        return detail::make_unique<algo_inner>(m_value);
     }
     // Mandatory methods.
     virtual population evolve(const population &pop) const override final
@@ -299,12 +253,16 @@ struct algo_inner final : algo_inner_base {
         pagmo_throw(not_implemented_error,
                     "The set_seed() method has been invoked but it is not implemented in the UDA");
     }
-    template <typename U, enable_if_t<pagmo::has_set_seed<U>::value && override_has_set_seed<U>::value, int> = 0>
+    template <typename U,
+              enable_if_t<detail::conjunction<pagmo::has_set_seed<U>, override_has_set_seed<U>>::value, int> = 0>
     static bool has_set_seed_impl(const U &a)
     {
         return a.has_set_seed();
     }
-    template <typename U, enable_if_t<pagmo::has_set_seed<U>::value && !override_has_set_seed<U>::value, int> = 0>
+    template <
+        typename U,
+        enable_if_t<detail::conjunction<pagmo::has_set_seed<U>, detail::negation<override_has_set_seed<U>>>::value,
+                    int> = 0>
     static bool has_set_seed_impl(const U &)
     {
         return true;
@@ -325,14 +283,16 @@ struct algo_inner final : algo_inner_base {
         pagmo_throw(not_implemented_error,
                     "The set_verbosity() method has been invoked but it is not implemented in the UDA");
     }
-    template <typename U,
-              enable_if_t<pagmo::has_set_verbosity<U>::value && override_has_set_verbosity<U>::value, int> = 0>
+    template <
+        typename U,
+        enable_if_t<detail::conjunction<pagmo::has_set_verbosity<U>, override_has_set_verbosity<U>>::value, int> = 0>
     static bool has_set_verbosity_impl(const U &a)
     {
         return a.has_set_verbosity();
     }
-    template <typename U,
-              enable_if_t<pagmo::has_set_verbosity<U>::value && !override_has_set_verbosity<U>::value, int> = 0>
+    template <typename U, enable_if_t<detail::conjunction<pagmo::has_set_verbosity<U>,
+                                                          detail::negation<override_has_set_verbosity<U>>>::value,
+                                      int> = 0>
     static bool has_set_verbosity_impl(const U &)
     {
         return true;
@@ -375,14 +335,28 @@ struct algo_inner final : algo_inner_base {
 
     // Serialization
     template <typename Archive>
-    void serialize(Archive &ar)
+    void serialize(Archive &ar, unsigned)
     {
-        ar(cereal::base_class<algo_inner_base>(this), m_value);
+        detail::archive(ar, boost::serialization::base_object<algo_inner_base>(*this), m_value);
     }
     T m_value;
 };
 
-} // end of namespace detail
+} // namespace detail
+
+} // namespace pagmo
+
+namespace boost
+{
+
+template <typename T>
+struct is_virtual_base_of<pagmo::detail::algo_inner_base, pagmo::detail::algo_inner<T>> : false_type {
+};
+
+} // namespace boost
+
+namespace pagmo
+{
 
 /// Algorithm class.
 /**
@@ -391,7 +365,7 @@ struct algo_inner final : algo_inner_base {
  * This class represents an optimization algorithm. An algorithm can be
  * stochastic, deterministic, population based, derivative-free, using hessians,
  * using gradients, a meta-heuristic, evolutionary, etc.. Via this class pagmo offers a common interface to
- * all types of algorithms that can be applied to find solution to a generic matematical
+ * all types of algorithms that can be applied to find solution to a generic mathematical
  * programming problem as represented by the pagmo::problem class.
  *
  * In order to define an optimizaztion algorithm in pagmo, the user must first define a class
@@ -426,27 +400,28 @@ struct algo_inner final : algo_inner_base {
  * \verbatim embed:rst:leading-asterisk
  * .. note::
  *
- *    A moved-from pagmo::algorithm is destructible and assignable. Any other operation will result
- *    in undefined behaviour.
+ *    The only operations allowed on a moved-from :cpp:class:`pagmo::algorithm` are destruction,
+ *    assignment, and the invocation of the :cpp:func:`~pagmo::algorithm::is_valid()` member function.
+ *    Any other operation will result in undefined behaviour.
  *
  * \endverbatim
  */
-class algorithm
+class PAGMO_DLL_PUBLIC algorithm
 {
     // Enable the generic ctor only if T is not an algorithm (after removing
     // const/reference qualifiers), and if T is a uda.
     template <typename T>
-    using generic_ctor_enabler
-        = enable_if_t<!std::is_same<algorithm, uncvref_t<T>>::value && is_uda<uncvref_t<T>>::value, int>;
+    using generic_ctor_enabler = enable_if_t<
+        detail::conjunction<detail::negation<std::is_same<algorithm, uncvref_t<T>>>, is_uda<uncvref_t<T>>>::value, int>;
+
+public:
+    // Default constructor.
+    algorithm();
+
+private:
+    void generic_ctor_impl();
 
 public:
-    /// Default constructor.
-    /**
-     * The default constructor will initialize a pagmo::algorithm containing a pagmo::null_algorithm.
-     *
-     * @throws unspecified any exception thrown by the constructor from UDA.
-     */
-    algorithm() : algorithm(null_algorithm{}) {}
     /// Constructor from a user-defined algorithm of type \p T
     /**
      * \verbatim embed:rst:leading-asterisk
@@ -471,71 +446,41 @@ public:
     template <typename T, generic_ctor_enabler<T> = 0>
     explicit algorithm(T &&x) : m_ptr(detail::make_unique<detail::algo_inner<uncvref_t<T>>>(std::forward<T>(x)))
     {
-        // We detect if set_seed is implemented in the algorithm, in which case the algorithm is stochastic
-        m_has_set_seed = ptr()->has_set_seed();
-        // We detect if set_verbosity is implemented in the algorithm
-        m_has_set_verbosity = ptr()->has_set_verbosity();
-        // We store at construction the value returned from the user implemented get_name
-        m_name = ptr()->get_name();
-        // Store the thread safety value.
-        m_thread_safety = ptr()->get_thread_safety();
+        generic_ctor_impl();
     }
-    /// Copy constructor
+    // Copy constructor
+    algorithm(const algorithm &);
+    // Move ctor.
+    algorithm(algorithm &&) noexcept;
+    // Move assignment.
+    algorithm &operator=(algorithm &&) noexcept;
+    // Copy assignment.
+    algorithm &operator=(const algorithm &);
+    /// Assignment from a user-defined algorithm of type \p T
     /**
-     * The copy constructor will deep copy the input algorithm \p other.
+     * \verbatim embed:rst:leading-asterisk
+     * .. note::
      *
-     * @param other the algorithm to be copied.
+     *    This operator is not enabled if, after the removal of cv and reference qualifiers,
+     *    ``T`` is of type :cpp:class:`pagmo::algorithm` (that is, this operator does not compete with the copy/move
+     *    assignment operators of :cpp:class:`pagmo::algorithm`), or if ``T`` does not satisfy
+     *    :cpp:class:`pagmo::is_uda`.
      *
-     * @throws unspecified any exception thrown by:
-     * - memory allocation errors in standard containers,
-     * - the copying of the internal UDA.
+     * \endverbatim
+     *
+     * This operator will set the internal UDA to ``x`` by constructing a pagmo::algorithm from ``x``, and then
+     * move-assigning the result to ``this``.
+     *
+     * @param x the UDA.
+     *
+     * @return a reference to ``this``.
+     *
+     * @throws unspecified any exception thrown by the constructor from UDA.
      */
-    algorithm(const algorithm &other)
-        : m_ptr(other.m_ptr->clone()), m_has_set_seed(other.m_has_set_seed),
-          m_has_set_verbosity(other.m_has_set_verbosity), m_name(other.m_name), m_thread_safety(other.m_thread_safety)
+    template <typename T, generic_ctor_enabler<T> = 0>
+    algorithm &operator=(T &&x)
     {
-    }
-    /// Move constructor
-    /**
-     * @param other the algorithm from which \p this will be move-constructed.
-     */
-    algorithm(algorithm &&other) noexcept
-        : m_ptr(std::move(other.m_ptr)), m_has_set_seed(std::move(other.m_has_set_seed)),
-          m_has_set_verbosity(other.m_has_set_verbosity), m_name(std::move(other.m_name)),
-          m_thread_safety(std::move(other.m_thread_safety))
-    {
-    }
-    /// Move assignment operator
-    /**
-     * @param other the assignment target.
-     *
-     * @return a reference to \p this.
-     */
-    algorithm &operator=(algorithm &&other) noexcept
-    {
-        if (this != &other) {
-            m_ptr = std::move(other.m_ptr);
-            m_has_set_seed = std::move(other.m_has_set_seed);
-            m_has_set_verbosity = other.m_has_set_verbosity;
-            m_name = std::move(other.m_name);
-            m_thread_safety = std::move(other.m_thread_safety);
-        }
-        return *this;
-    }
-    /// Copy assignment operator
-    /**
-     * Copy assignment is implemented as a copy constructor followed by a move assignment.
-     *
-     * @param other the assignment target.
-     *
-     * @return a reference to \p this.
-     *
-     * @throws unspecified any exception thrown by the copy constructor.
-     */
-    algorithm &operator=(const algorithm &other)
-    {
-        // Copy ctor + move assignment.
-        return *this = algorithm(other);
+        return (*this) = algorithm(std::forward<T>(x));
     }
 
     /// Extract a const pointer to the UDA.
@@ -557,7 +502,7 @@ public:
      * in the constructor.
      */
     template <typename T>
-    const T *extract() const
+    const T *extract() const noexcept
     {
         auto p = dynamic_cast<const detail::algo_inner<T> *>(ptr());
         return p == nullptr ? nullptr : &(p->m_value);
@@ -590,7 +535,7 @@ public:
      * in the constructor.
      */
     template <typename T>
-    T *extract()
+    T *extract() noexcept
     {
         auto p = dynamic_cast<detail::algo_inner<T> *>(ptr());
         return p == nullptr ? nullptr : &(p->m_value);
@@ -601,42 +546,16 @@ public:
      * @return \p true if the user-defined algorithm is \p T, \p false otherwise.
      */
     template <typename T>
-    bool is() const
+    bool is() const noexcept
     {
         return extract<T>() != nullptr;
     }
 
-    /// Evolve method.
-    /**
-     * This method will invoke the <tt>%evolve()</tt> method of the UDA. This is where the core of the optimization
-     * (*evolution*) is made.
-     *
-     * @param pop starting population
-     *
-     * @return evolved population
-     *
-     * @throws unspecified any exception thrown by the <tt>%evolve()</tt> method of the UDA.
-     */
-    population evolve(const population &pop) const
-    {
-        return ptr()->evolve(pop);
-    }
+    // Evolve method.
+    population evolve(const population &) const;
 
-    /// Set the seed for the stochastic evolution.
-    /**
-     * Sets the seed to be used in the <tt>%evolve()</tt> method of the UDA for all stochastic variables. If the UDA
-     * satisfies pagmo::has_set_seed, then its <tt>%set_seed()</tt> method will be invoked. Otherwise, an error will be
-     * raised.
-     *
-     * @param seed seed.
-     *
-     * @throws not_implemented_error if the UDA does not satisfy pagmo::has_set_seed.
-     * @throws unspecified any exception thrown by the <tt>%set_seed()</tt> method of the UDA.
-     */
-    void set_seed(unsigned seed)
-    {
-        ptr()->set_seed(seed);
-    }
+    // Set the seed for the stochastic evolution.
+    void set_seed(unsigned);
 
     /// Check if a <tt>%set_seed()</tt> method is available in the UDA.
     /**
@@ -665,22 +584,8 @@ public:
         return has_set_seed();
     }
 
-    /// Set the verbosity of logs and screen output.
-    /**
-     * This method will set the level of verbosity for the algorithm. If the UDA satisfies pagmo::has_set_verbosity,
-     * then its <tt>%set_verbosity()</tt> method will be invoked. Otherwise, an error will be raised.
-     *
-     * The exact meaning of the input parameter \p level is dependent on the UDA.
-     *
-     * @param level the desired verbosity level.
-     *
-     * @throws not_implemented_error if the UDA does not satisfy pagmo::has_set_verbosity.
-     * @throws unspecified any exception thrown by the <tt>%set_verbosity()</tt> method of the UDA.
-     */
-    void set_verbosity(unsigned level)
-    {
-        ptr()->set_verbosity(level);
-    }
+    // Set the verbosity of logs and screen output.
+    void set_verbosity(unsigned);
 
     /// Check if a <tt>%set_verbosity()</tt> method is available in the UDA.
     /**
@@ -715,19 +620,8 @@ public:
         return m_name;
     }
 
-    /// Algorithm's extra info.
-    /**
-     * If the UDA satisfies pagmo::has_extra_info, then this method will return the output of its
-     * <tt>%get_extra_info()</tt> method. Otherwise, an empty string will be returned.
-     *
-     * @return extra info about the UDA.
-     *
-     * @throws unspecified any exception thrown by the <tt>%get_extra_info()</tt> method of the UDA.
-     */
-    std::string get_extra_info() const
-    {
-        return ptr()->get_extra_info();
-    }
+    // Algorithm's extra info.
+    std::string get_extra_info() const;
 
     /// Algorithm's thread safety level.
     /**
@@ -742,33 +636,8 @@ public:
         return m_thread_safety;
     }
 
-    /// Streaming operator
-    /**
-     * This function will stream to \p os a human-readable representation of the input
-     * algorithm \p a.
-     *
-     * @param os input <tt>std::ostream</tt>.
-     * @param a pagmo::algorithm object to be streamed.
-     *
-     * @return a reference to \p os.
-     *
-     * @throws unspecified any exception thrown by querying various algorithm properties and streaming them into \p os.
-     */
-    friend std::ostream &operator<<(std::ostream &os, const algorithm &a)
-    {
-        os << "Algorithm name: " << a.get_name();
-        if (!a.has_set_seed()) {
-            stream(os, " [deterministic]");
-        } else {
-            stream(os, " [stochastic]");
-        }
-        stream(os, "\n\tThread safety: ", a.get_thread_safety(), '\n');
-        const auto extra_str = a.get_extra_info();
-        if (!extra_str.empty()) {
-            stream(os, "\nExtra info:\n", extra_str);
-        }
-        return os;
-    }
+    // Check if the algorithm is valid.
+    bool is_valid() const;
 
     /// Save to archive.
     /**
@@ -779,9 +648,9 @@ public:
      * @throws unspecified any exception thrown by the serialization of the UDA and of primitive types.
      */
     template <typename Archive>
-    void save(Archive &ar) const
+    void save(Archive &ar, unsigned) const
     {
-        ar(m_ptr, m_has_set_seed, m_has_set_verbosity, m_name, m_thread_safety);
+        detail::to_archive(ar, m_ptr, m_has_set_seed, m_has_set_verbosity, m_name, m_thread_safety);
     }
     /// Load from archive.
     /**
@@ -792,12 +661,14 @@ public:
      * @throws unspecified any exception thrown by the deserialization of the UDA and of primitive types.
      */
     template <typename Archive>
-    void load(Archive &ar)
+    void load(Archive &ar, unsigned)
     {
         algorithm tmp;
-        ar(tmp.m_ptr, tmp.m_has_set_seed, tmp.m_has_set_verbosity, tmp.m_name, tmp.m_thread_safety);
+        detail::from_archive(ar, tmp.m_ptr, tmp.m_has_set_seed, tmp.m_has_set_verbosity, tmp.m_name,
+                             tmp.m_thread_safety);
         *this = std::move(tmp);
     }
+    BOOST_SERIALIZATION_SPLIT_MEMBER()
 
 private:
     // Two small helpers to make sure that whenever we require
@@ -824,8 +695,13 @@ private:
     std::string m_name;
     thread_safety m_thread_safety;
 };
+
+// Streaming operator for algorithm.
+PAGMO_DLL_PUBLIC std::ostream &operator<<(std::ostream &, const algorithm &);
+
 } // namespace pagmo
 
-PAGMO_REGISTER_ALGORITHM(pagmo::null_algorithm)
+// Disable tracking for the serialisation of algorithm.
+BOOST_CLASS_TRACKING(pagmo::algorithm, boost::serialization::track_never)
 
 #endif
